/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/


// camera.h
// header files for camera manipulation

#ifndef ALPSHOOT_CAMERA_H
#define ALPSHOOT_CAMERA_H

#include <string.h>

#ifdef ALPSHOOT_WIN32
#include <windows.h>
#endif

#ifndef ALPSHOOT_MACOSX
#include <GL/glut.h>
#else
#include <glut.h>
#endif

#include "world.h"
#include "quatlib.h"

/* camera structure */
struct camera_t {
	// quaternion that specifies the current orientation
	quaternion orientation;
	// camera radius (useful for collision detection)
	GLdouble radius;
	// total matrix (specifies position and orientation of the camera)
	GLdouble tot[16];
	// camera position in the previous frame
	vector old_pos;
	// how much to shift along the X axis next frame
	GLdouble to_shift_x;
	// how much to shift along the Y axis next frame
	GLdouble to_shift_y;
	// how much to shift along the Z axis next frame
	GLdouble to_shift_z;
	// camera position
	vector pos;
	// camera viewpoint (Z axis)
	vector view;
	// camera up vector (Y axis)
	vector up;
	// camera right (strafe) vector (X axis)
	vector right;
	// total rotation around X axis of main coordinate system
	GLdouble total_rot_x;
	// total rotation around Y axis of main coordinate system
	GLdouble total_rot_y;
	// total rotation around Z axis of main coordinate system
	GLdouble total_rot_z;
	// specifies whether camera is auto-leveling
	int leveling;
};
typedef struct camera_t camera_t;

void position_camera(GLdouble posx, GLdouble posy, GLdouble posz, GLdouble viewx, GLdouble viewy, GLdouble viewz, GLdouble upx, GLdouble upy, GLdouble upz);
void reset_camera();
void update_camera();
void check_camera_collision(vector *v, long v_count, GLdouble t[16]);
void rotate_around_X(GLdouble angle, camera_t *cam);
void rotate_around_Y(GLdouble angle, camera_t *cam);
void rotate_around_Z(GLdouble angle, camera_t *cam);
void auto_level(camera_t *cam);
int get_parallel(vector target, vector local_axis, camera_t *cam);

extern camera_t* cam;
extern camera_t* outcam;

#endif
