/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/


// display.c
// display callback functions

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "display.h"


// Object list functions

/* Draws targets */
void draw_target()
{
	GLfloat old_diffuse[4];
	GLfloat ps, ls;
	
	lights_off();
	
	glPushMatrix();
	
	if (target->live == 2)
	{
		glCallList(dl->green_list);
		glGetMaterialfv(GL_FRONT, GL_DIFFUSE, old_diffuse);
		if (fps->fps > 1.0)
			target->alpha -= 0.002*100.0/fps->fps;
		else
			target->alpha -= 0.002;
		old_diffuse[3] = target->alpha;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, old_diffuse);
	}
	else
	{
		glCallList(dl->red_list);
		//glGetMaterialfv(GL_FRONT, GL_DIFFUSE, old_diffuse);
		//old_diffuse[0] += 3*target->t;
		//old_diffuse[1] -= 3*target->t;
		//glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, old_diffuse);
	}
	
	glTranslated(target->vp.x, target->vp.y, target->vp.z);
	glRotated(-target->rotate_around_z, 0.0, 0.0, 1.0);
	glRotated(-target->rotate_around_y, 0.0, 1.0, 0.0);
	glRotated(-target->rotate_around_x, 1.0, 0.0, 0.0);
	
	gluQuadricTexture(qobj, GL_FALSE);
	gluQuadricOrientation(qobj, GLU_INSIDE);
	gluQuadricDrawStyle(qobj, GLU_LINE);
	gluQuadricNormals(qobj, GLU_NONE);

	gluSphere(qobj, target->radius, 20, 20);
	
	glPopMatrix();

	if (state->lines)
	{
		glBegin(GL_LINES);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x+target->radius, target->vp.y, target->vp.z);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x-target->radius, target->vp.y, target->vp.z);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x, target->vp.y, target->vp.z+target->radius);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x, target->vp.y, target->vp.z-target->radius);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x, target->vp.y+target->radius, target->vp.z);
		glVertex3f(0.0, 0.125, 20.0);
		glVertex3f(target->vp.x, target->vp.y-target->radius, target->vp.z);
		glEnd();
	}
	if (state->trajectories)
	{
		glGetFloatv(GL_LINE_WIDTH, &ls);
		glLineWidth(2.0);
		glMap1d(GL_MAP1_VERTEX_4, 0.0, 1.0, 4, 4, &target->init_vp.x);
		glEnable(GL_MAP1_VERTEX_4);
		glMapGrid1d(40, 0.0, 1.0);
		glEvalMesh1(GL_LINE, 0.0, 40);
		glDisable(GL_MAP1_VERTEX_4);
		glLineWidth(ls);
		
		glGetFloatv(GL_POINT_SIZE, &ps);
		glPointSize(5.0);
		glBegin(GL_POINTS);
		glVertex3f(target->init_vp.x, target->init_vp.y, target->init_vp.z);
		glVertex3f(target->cp1.x, target->cp1.y, target->cp1.z);
		glVertex3f(target->cp2.x, target->cp2.y, target->cp2.z);
		glVertex3f(target->final_vp.x, target->final_vp.y, target->final_vp.z);
		glEnd();
		glPointSize(ps);
	}
	
	if (state->light0 == 1)
		lights_on();
	glCallList(dl->white_list);
}

/* Draws a crate */
void draw_crate()
{
	glCallList(dl->crate_list);
}

/* Draws a barrel */
void draw_barrel()
{
	glCallList(dl->barrel_list);
}

/* Enables a vanishing effect */
void vanish(object_t* op)
{
	op->live = 2;
}

/* Draws an explosion effect */
void explode(object_t* op)
{
	new_explosion(1, op->vp);
	op->live = 0;
}

/* Creates objects to draw */
void fill_object_list()
{
	object_t *tmp = NULL;
	
	// crates
	// right
	make_object(1.5, 0.0, 185.0, 0.0, -12.8, 1.0, -15.0, &draw_crate, &explode); 
	make_object(1.5, 0.0, -10.0, 0.0, -13.0, 1.0, -17.3, &draw_crate, &explode);
	make_object(1.5, 0.0, 90.0, 0.0, -15.0, 1.0, -15.0, &draw_crate, &explode);
	make_object(1.5, 0.0, 110.0, 0.0, -14.0, 3.0, -14.8, &draw_crate, &explode);
	//top
	make_object(1.5, 0.0, 110.0, 0.0, 0.0, 1.0, 28.0, &draw_crate, &explode);
	make_object(1.5, 0.0, 0.0, 90.0, -3.0, 1.0, 28.0, &draw_crate, &explode);
	make_object(1.5, 0.0, 0.0, 0.0, -20.0, 1.0, 25.0, &draw_crate, &explode);
	make_object(1.5, 90.0, 0.0, 0.0, 21.0, 1.0, 25.0, &draw_crate, &explode);
	make_object(1.5, 0.0, 90.0, 0.0, -20.0, 1.0, 10.0, &draw_crate, &explode);
	// left
	make_object(1.5, 0.0, 30.0, 0.0, 16.0, 1.0, -27.0, &draw_crate, &explode);
	make_object(1.5, 0.0, 50.0, 0.0, 20.0, 1.0, -26.5, &draw_crate, &explode);
	make_object(1.5, 0.0, 90.0, 90.0, 22.0, 1.0, -6.5, &draw_crate, &explode);
	// barrels
	// left
	tmp = make_object(1.7, -90.0, 0.0, 0.0, 13.0, 0.1, -19.0, &draw_barrel, &explode);
	tmp->vp.y += 1.1;
	tmp = make_object(1.7, 0.0, 0.0, 110.0, 11.5, 0.9, -22.0, &draw_barrel, &explode);
	tmp->vp.y -= 1.1;
	tmp = make_object(1.7, -90.0, 0.0, 120.0, 16.5, 0.1, -22.0, &draw_barrel, &explode);
	tmp->vp.y += 1.1;
	tmp = make_object(1.7, -90.0, 0.0, -30.0, 16.45, 0.1, -20.0, &draw_barrel, &explode);
	tmp->vp.y += 1.1;
	// right
	tmp = make_object(1.7, -90.0, 0.0, -30.0, -15.0, 0.0, -17.5, &draw_barrel, &explode);
	tmp->vp.y += 1.1;
	tmp = make_object(1.7, 0.0, -30.0, 0.0, -17.0, 0.8, -17.0, &draw_barrel, &explode);
	tmp->vp.z += 1.1;
	tmp = make_object(1.7, -46.0, 0.0, 180.0, -16.6, 0.5, -19.0, &draw_barrel, &explode);
	tmp->vp.y += 1.1;
	tmp = make_object(1.7, 0.0, 20.0, 90.0, -14.0, 0.8, -28.0, &draw_barrel, &explode);
	tmp->vp.z += 1.1;
	// target placeholder
	tmp = add_object();
	target = tmp;
}

/* Updates shooter targets */
void update_shooter()
{
	if (state->shooter > 0)
	{
		// move target
		if (target->live > 0)
		{
			if (target->t <= 1.0)
			{
				target->vp = de_cast4(target->init_vp, target->cp1, target->cp2, target->final_vp, target->t);
				if (fps->fps > 1.0)
					target->t += 0.0015*100.0/fps->fps;
				else
					target->t += 0.0015;
			}
			else
			{
				target->live = 0;
				target->t = 0.0;
			}
		}
		// create new target
		else
		{
			state->shooter++;
			if (state->shooter == 20)
			{
				state->light0 = 1;
				lights_on();
				state->shooter = 0;
			}
			else
			{
				target->radius = 1.5;
				// a+drand*(b-a)
				target->vp.x = -25.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*50.0;
				target->vp.y = 5.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*8.0;
				target->vp.z = 8.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*24.0;
				target->vp.w = 1.0;
				copy_vector(target->vp, &(target->init_vp));
				if (target->vp.x > 0)
					target->final_vp.x = -25.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*25.0;
				else
					target->final_vp.x = 25.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*25.0;
				target->final_vp.y = 3.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*10.0;
				target->final_vp.z = 10.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*22.0;
				target->final_vp.w = 1.0;
				target->cp1.x = -25.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*50.0;
				target->cp1.y = 5.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*8.0;
				target->cp1.z = 8.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*24.0;
				target->cp1.w = 1.0;
				target->cp2.x = -25.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*50.0;
				target->cp2.y = 5.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*8.0;
				target->cp2.z = 8.0+((GLdouble)rand()/(GLdouble)RAND_MAX)*24.0;
				target->cp2.w = 1.0;
				target->draw = &draw_target;
				target->on_collision = &vanish;
				target->t = 0.0;
				target->alpha = 0.5;
				target->live = 1;
			}
		}
	}
}

// Drawing functions

/* Draws rotating planet and moon */
void draw_planet()
{
	glPushMatrix();

	glEnable(GL_TEXTURE_2D);
	
	glDisable(GL_LIGHT1);
	lights_off();
	
	// draw planet
	glTranslated(-10.0, 2.5, 55.0);
	glRotated(-70.0, 1.0, 0.0, 0.0);
	glRotated(90.0, 0.0, 0.0, 1.0);
	if (fps->fps > 1.0)
		state->pl_ang += 0.005*100.0/fps->fps;
	else
		state->pl_ang += 0.005;
	if (state->pl_ang >= 360.0)
		state->pl_ang = 0.0;
	glRotated(state->pl_ang, 0.0, 0.0, 1.0);

	glBindTexture(GL_TEXTURE_2D, tex->tex_names[BRONZE]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	gluQuadricTexture(qobj, GL_TRUE);
	gluQuadricOrientation(qobj, GLU_INSIDE);
	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);

	gluSphere(qobj, 13.0, 30, 30);

	// draw moon
	glRotated(2*state->pl_ang, 0.0, 0.0, 1.0);
	glTranslated(0.0, -16.0, 0.0);
	
	glBindTexture(GL_TEXTURE_2D, tex->tex_names[METAL]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	
	gluQuadricTexture(qobj, GL_TRUE);
	gluQuadricOrientation(qobj, GLU_INSIDE);
	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);
	
	gluSphere(qobj, 1.0, 20, 20);

	glPopMatrix();

	glDisable(GL_TEXTURE_2D);

	if (state->light0 == 1)
		lights_on();
	if (state->light1)
		glEnable(GL_LIGHT1);
}

/* Draws a compass */
/* This is done by calculating the angle between the projection of the current 
 * camera view vector on the XZ plane and the Z axis, then rotating the compass
 * accordingly */
void draw_compass()
{
	const vector z_axis = { 0.0, 0.0, 1.0, 1.0 }, x_axis = { 1.0, 0.0, 0.0, 1.0 }, cam_view_proj = { cam->view.x-cam->pos.x, 0.0, cam->view.z-cam->pos.z, 1.0 };
	vector cam_view_temp = { dot_product(cam_view_proj, x_axis), 0.0, dot_product(cam_view_proj, z_axis), 1.0 };
	GLdouble angle = 0.0;
	
	glPushMatrix();
	glLoadIdentity();
	glTranslated(0.0, 0.5, -1.0);
	glRotated(115.0, 1.0, 0.0, 0.0);
	glRotated(10.0, 0.0, 0.0, 1.0);
	
	cam_view_temp = normalize(cam_view_temp);	
	// angle = acos(dp/m1*m2)
	// m2 being 1.0 here and dp being trivial
	angle = 180.0*((GLdouble)acos(cam_view_temp.z/magnitude(cam_view_temp)))/M_PI;
	if (isnan(angle))
		angle = 0.0;
	
	// adjust angle depending on what quadrant you're looking at
	if (cam_view_temp.x > 0.0)
		angle = 360.0-angle;
	
	// rotating around Z axis is due to the way dlist is composed
	glRotated(angle, 0.0, 0.0, 1.0);
	glCallList(dl->compass_list);
	glPopMatrix();
}

/* Writes text on screen in the specified position and color */
void draw_text2d(GLint x, GLint y, char *string, GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	GLfloat emit[] = { r, g, b, 1.0 };
	GLfloat diff[] = { 0.0 , 0.0, 0.0, a };
	GLint len = 0, i = 0;

	glMaterialfv(GL_FRONT, GL_EMISSION, emit);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, diff);
	
	/* draw text */
	glRasterPos2i(x, y);
	
	len = (int)strlen(string);
	
	for(i = 0; i < len; i++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, string[i]);
}

/* Draws the mouse cursor position */
void draw_mouse()
{
	const GLfloat cursor_emit_red[] = { 0.5, 0.0, 0.0, 1.0 };
	const GLfloat cursor_emit_brightred[] = { 0.7, 0.0, 0.0, 1.0 };
	GLfloat old_size = 0.0;
	GLint posx = state->mouse_posx, posy = state->mouse_posy, winx = state->windowx/2.0, winy = state->windowy/2.0;
	GLint hor_proj = 0, ver_proj = 0;
	GLdouble cen_proj = 0.0, cos_angle = 0.0, angle = 0.0;
	const GLfloat CURSOR_SCALE = 20.0;

	// if in persistent mode there could be some work to do
	if (state->mouse_mode == MOUSE_PERSISTENT)
	{
		glPushMatrix();
		glTranslatef(posx, posy, 0.0);
		// if in dead zone, draw a point
		if ((state->xprop == 0.0) && (state->yprop == 0.0))
		{	
			glGetFloatv(GL_POINT_SIZE, &old_size);
			glMaterialfv(GL_FRONT, GL_EMISSION, cursor_emit_red);
			glPointSize(8.0);
			glBegin(GL_POINTS);
			glVertex2f((GLfloat)0.0, (GLfloat)0.0);
			glEnd();
			glPointSize(old_size);
		}
		// else draw direction triangle
		else
		{
			hor_proj = posx - winx;
			ver_proj = posy - winy;
			cen_proj = (GLdouble)sqrt((double)(hor_proj*hor_proj + ver_proj*ver_proj));
			cos_angle = cen_proj <  -F_PRECISION || cen_proj > F_PRECISION ? (GLdouble)(hor_proj/(GLdouble)cen_proj) : (GLdouble)hor_proj;
			angle = 180.0*acos(cos_angle)/M_PI;
			if (posy > winy)
				glRotatef(angle, 0.0, 0.0, 1.0);
			else
				glRotatef(-angle, 0.0, 0.0, 1.0);
			glMaterialfv(GL_FRONT, GL_EMISSION, cursor_emit_red);
			glBegin(GL_TRIANGLES);
			glVertex2f(2.0, 2.0);
			glVertex2f(8.0+cen_proj/CURSOR_SCALE, 5.0);
			glVertex2f(2.0, 8.0);
			glEnd();
			glMaterialfv(GL_FRONT, GL_EMISSION, cursor_emit_brightred);
			glBegin(GL_TRIANGLES);
			glVertex2f(0.0, 0.0);
			glVertex2f(10.0+cen_proj/CURSOR_SCALE, 5.0);
			glVertex2f(0.0, 10.0);
			glEnd();
		}
		glPopMatrix();
	}
	// else draw a point in the center of the screen
	else
	{
		glMaterialfv(GL_FRONT, GL_EMISSION, cursor_emit_red);
		glGetFloatv(GL_POINT_SIZE, &old_size);
		glPointSize(4.0);
		glBegin(GL_POINTS);
		glVertex2f((GLfloat)winx, (GLfloat)winy);
		glEnd();
		glPointSize(old_size);
	}
}

/* Enters othographics projection mode to draw text or images to the
 * front of the screen */
void enter_ortho_mode()
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(0.0, (GLdouble)state->windowx, (GLdouble)state->windowy, 0.0, 0.0, 1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
}

/* Gets back to perspective projection mode after entering orthographic mode */
void exit_ortho_mode()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

/* Draws the heads up display on screen */
void draw_hud()
{
	GLfloat old_diffuse[4];

	enter_ortho_mode();

	glGetMaterialfv(GL_FRONT, GL_DIFFUSE, old_diffuse);
	
	draw_text2d(20, 30, "Score:", 0.0, 1.0, 0.0, 1.0);
	draw_text2d(60, 30, state->score, 0.0, 1.0, 0.0, 1.0);
	if (left_shot->t > 0)
		draw_text2d(20, 50, "Ready", 1.0, 0.0, 0.0, 1.0);
	else
		draw_text2d(20, 50, "Ready", 0.0, 1.0, 0.0, 1.0);
	if (right_shot->t > 0)
		draw_text2d(state->windowx - 60, 50, "Ready", 1.0, 0.0, 0.0, 1.0);
	else
		draw_text2d(state->windowx - 60, 50, "Ready", 0.0, 1.0, 0.0, 1.0);
	
	if (state->cannons_linked)
		draw_text2d(20, 70, "Dual-linked", 0.0, 1.0, 0.0, 1.0);

	if (state->mouse_mode == MOUSE_PROGRESSIVE)
		draw_text2d(state->windowx-130, 70, "Progressive mouse", 0.0, 1.0, 0.0, 1.0);
	else if (state->mouse_mode == MOUSE_PERSISTENT)
		draw_text2d(state->windowx-120, 70, "Persistent mouse", 0.0, 1.0, 0.0, 1.0);

	if (state->jetpack)
		draw_text2d((state->windowx/2)-33, 30, "Jetpack on", 0.0, 1.0, 0.0, 1.0);
		
	if (strlen(state->status_string))
		draw_text2d(state->windowx/2 - 40, state->windowy - 40, state->status_string, 0.0, 1.0, 0.0, 1.0);

	/* draw framerate */
	draw_text2d(state->windowx-(66+strlen(fps->fps_string)), 30, fps->fps_string, 0.0, 1.0, 0.0, 1.0);
	
	draw_mouse();
	/* crosshair */
	draw_crosshair();
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, old_diffuse);

	exit_ortho_mode();	
}

/* Determines whether a rocket hits solid boundaries or not */
int shot_hit(vector v)
{
	if ((v.y < 0.0) || (v.y > 16.0))
		return 1;
	else if ((v.x > 23.0) || (v.x < -23.0))
		return 1;
	else if ((v.z > 30.0) || (v.z < -30.0))
		return 1;
	// monitors
	else if ((v.z < 21.0) && (v.z > 13.0))
	{
		if ((v.y > 0.0) && (v.y <= 4.0))
		{
			if ((v.x < -20.0) || (v.x > 20.0))
				return 1;
		}
		else if ((v.y > 4.0) && (v.y <= 5.0))
		{
			if ((v.x < -20.5) || (v.x > 20.5))
				return 1;
		}
		else if ((v.y > 5.0) && (v.y <= 6.0))
		{
			if ((v.x < -21.5) || (v.x > 21.5))
				return 1;
		}
		else if ((v.y > 6.0) && (v.y <= 7.0))
		{
			if ((v.x < -22.5) || (v.x > 22.5))
				return 1;
		}
	}
	// fence
	else if ((v.z <= -8.65) && (v.z >= -31.0))
	{
		if ((v.y > 0.0) && (v.y <= 4.0))
		{
			if ((v.x <= -8.85) && (v.x >= -9.85))
				return 1;
			else if ((v.x < -8.85) && (v.x >= -20.0))
			{
				if ((v.z <= -8.65) && (v.z >= -9.65))
					return 1;
			}
		}
	}
	return 0;
}

/* Draws cannon shots */
/* These are currently rockets but could be anything */
void draw_shots(object_t *shot)
{
	vector axis;
	GLdouble angle;
	
	glCallList(dl->shots_list);
	
	if ((shot->live) || (shot->t > 0.0))
	{
		if (shot->t <= 1.0)
		{
			// find new shot position for this frame
			shot->vp = de_cast2(shot->init_vp, shot->final_vp, shot->t);
			// see if shots should explode this frame
			if (shot->live && shot_hit(shot->vp))
			{
				shot->live = 0;
				explode(shot);
			}
			// otherwise draw it
			if (shot->live)
			{
				glPushMatrix();
				glTranslated(shot->vp.x, shot->vp.y, shot->vp.z);
				/*
				glRotated(-shot->rotate_around_z, 0.0, 0.0, 1.0);
				glRotated(-shot->rotate_around_y, 0.0, 1.0, 0.0);
				glRotated(-shot->rotate_around_x, 1.0, 0.0, 0.0);
				*/
				axis = q_convert_aa(shot->orientation, &angle);
				angle = -angle*180.0/M_PI;
				glRotated(angle, axis.x, axis.y, axis.z);
				glDisable(GL_CULL_FACE);
				glmDraw(rocket, GLM_SMOOTH | GLM_MATERIAL);
				glEnable(GL_CULL_FACE);
				glPopMatrix();
				new_trail(1, shot->vp);
			}
			// 0.003
			if (fps->fps > 1.0)
				shot->t += 0.005*100.0/fps->fps;
			else
				shot->t += 0.005;
		}
		// explode at end of fuel
		else
		{
			shot->t = 0.0;
			if (shot->live)
				explode(shot);
			shot->live = 0;
		}
	}	
}

/* Draws rocket trails */
void draw_trails()
{
	int i = 0;
	trail_t *tmp = trail_list;
	trail_t *nexttmp = NULL;
	
	glCallList(dl->white_list);

	while (tmp != NULL)
	{
		// draw particles
		glPushMatrix();

		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);

		glBegin(GL_POINTS);

		for (i = 0; i < NUM_TRAIL; i++)
		{
			glColor4fv(tmp->particles[i].color);
			glVertex3fv(tmp->particles[i].position);
		}

		glEnd();

		glPopMatrix();

		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);

		// update particles status
		for (i = 0; i < NUM_TRAIL; i++)
		{
			tmp->particles[i].position[0] += tmp->particles[i].speed[0] * 0.005;
			tmp->particles[i].position[1] += tmp->particles[i].speed[1] * 0.005;
			tmp->particles[i].position[2] += tmp->particles[i].speed[2] * 0.005;

			tmp->particles[i].color[0] -= 1.0 / 500.0;
			if (tmp->particles[i].color[0] < 0.0)
			{
				tmp->particles[i].color[0] = 0.0;
			}

			tmp->particles[i].color[1] -= 1.0 / 100.0;
			if (tmp->particles[i].color[1] < 0.0)
			{
				tmp->particles[i].color[1] = 0.0;
			}

			tmp->particles[i].color[2] -= 1.0 / 50.0;
			if (tmp->particles[i].color[2] < 0.0)
			{
				tmp->particles[i].color[2] = 0.0;
			}

			tmp->particles[i].color[3] -= 1.0 / 50.0;
			if (tmp->particles[i].color[3] < 0.0)
			{
				tmp->particles[i].color[3] = 0.0;
			}
		}

		--(tmp->fuel);

		if (tmp->fuel == 0)
		{
			nexttmp = tmp->next;
			delete_trail(tmp);
			tmp = nexttmp;
		}
		else
			tmp = tmp->next;
	}
}

/* Draws pending explosions */
void draw_explosions()
{
	int i = 0;
	explosion_t *tmp = exp_list;
	explosion_t *nexttmp = NULL;
	
	glCallList(dl->explosion_list);

	while (tmp != NULL)
	{
		// draw particles
		glPushMatrix();

		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);

		glBegin(GL_POINTS);

		for (i = 0; i < NUM_PARTICLES; i++)
		{
			glColor4fv(tmp->particles[i].color);
			glVertex3fv(tmp->particles[i].position);
		}

		glEnd();

		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);

		glPopMatrix();

		glNormal3f (0.0, 0.0, 1.0);

		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		for (i = 0; i < NUM_DEBRIS; i++)
		{
			glColor4fv(tmp->debris[i].color);

			glPushMatrix();

			glTranslatef(tmp->debris[i].position[0],
					tmp->debris[i].position[1],
					tmp->debris[i].position[2]);

			glRotatef(tmp->debris[i].orientation[0], 1.0, 0.0, 0.0);
			glRotatef(tmp->debris[i].orientation[1], 0.0, 1.0, 0.0);
			glRotatef(tmp->debris[i].orientation[2], 0.0, 0.0, 1.0);

			glScalef(tmp->debris[i].scale[0],
					tmp->debris[i].scale[1],
					tmp->debris[i].scale[2]);

			glBegin(GL_TRIANGLES);
			glVertex3f(0.0, 0.5, 0.0);
			glVertex3f(-0.25, 0.0, 0.0);
			glVertex3f(0.25, 0.0, 0.0);
			glEnd();

			glPopMatrix();
		}
		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);

		glShadeModel(GL_SMOOTH);

		// update particles status
		for (i = 0; i < NUM_PARTICLES; i++)
		{
			tmp->particles[i].position[0] += tmp->particles[i].speed[0] * 0.2;
			tmp->particles[i].position[1] += tmp->particles[i].speed[1] * 0.2;
			tmp->particles[i].position[2] += tmp->particles[i].speed[2] * 0.2;

			tmp->particles[i].color[0] -= 1.0 / 500.0;
			if (tmp->particles[i].color[0] < 0.0)
			{
				tmp->particles[i].color[0] = 0.0;
			}

			tmp->particles[i].color[1] -= 1.0 / 100.0;
			if (tmp->particles[i].color[1] < 0.0)
			{
				tmp->particles[i].color[1] = 0.0;
			}

			tmp->particles[i].color[2] -= 1.0 / 50.0;
			if (tmp->particles[i].color[2] < 0.0)
			{
				tmp->particles[i].color[2] = 0.0;
			}

			tmp->particles[i].color[3] -= 1.0 / 50.0;
			if (tmp->particles[i].color[3] < 0.0)
			{
				tmp->particles[i].color[3] = 0.0;
			}
		}

		for (i = 0; i < NUM_DEBRIS; i++)
		{
			tmp->debris[i].position[0] += tmp->debris[i].speed[0] * 0.1;
			tmp->debris[i].position[1] += tmp->debris[i].speed[1] * 0.1;
			tmp->debris[i].position[2] += tmp->debris[i].speed[2] * 0.1;

			tmp->debris[i].orientation[0] += tmp->debris[i].orientationSpeed[0] * 10;
			tmp->debris[i].orientation[1] += tmp->debris[i].orientationSpeed[1] * 10;
			tmp->debris[i].orientation[2] += tmp->debris[i].orientationSpeed[2] * 10;

			tmp->debris[i].color[3] -= 1.0 / 50.0;
			if (tmp->debris[i].color[3] < 0.0)
			{
				tmp->debris[i].color[3] = 0.0;
			}
		}	
		
		--(tmp->fuel);
		
		if (tmp->fuel == 0)
		{
			nexttmp = tmp->next;
			delete_explosion(tmp);
			tmp = nexttmp;
		}
		else
			tmp = tmp->next;
	}
}

/* Draws control console */
void draw_console()
{
	glCallList(dl->console_list);
	
	glPushMatrix();
	
	glTranslated(5.0, 1.8, 10.0);
	glRotated(140.0, 1.0, 0.0, 0.0);

	glEnable(GL_TEXTURE_2D);
	
	if (state->shooter)
		glBindTexture(GL_TEXTURE_2D, tex->tex_names[CONSOLE]);
	else
		glBindTexture(GL_TEXTURE_2D, tex->tex_names[CONSOLE2]);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0); glVertex3f(1.2, 0.0, 0.8);
	glTexCoord2f(0.0, 1.0); glVertex3f(1.2, 0.0, -0.8);
	glTexCoord2f(1.0, 1.0); glVertex3f(-1.2, 0.0, -0.8);
	glTexCoord2f(1.0, 0.0); glVertex3f(-1.2, 0.0, 0.8);

	glEnd();
	
	glDisable(GL_TEXTURE_2D);
	lights_off();
	
	if (state->shooter)
	{	
		glTranslated(0.0, -0.8, 0.0);

		glCallList(dl->green_list);

		gluQuadricTexture(qobj, GL_FALSE);
		gluQuadricOrientation(qobj, GLU_INSIDE);
		gluQuadricDrawStyle(qobj, GLU_LINE);
		gluQuadricNormals(qobj, GLU_NONE);

		glRotated(50*(state->pl_ang), 0.0, 1.0, 0.0);
		gluSphere(qobj, 0.3, 20, 20);

		glCallList(dl->white_list);
	}

	if (state->light0 == 1)
		lights_on();
	
	glPopMatrix();
}

/* Draws lights on the ceiling */
void draw_lights()
{
	GLfloat emit[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat specular[] = { 0.1, 0.1, 0.1, 1.0 };
	GLfloat mat_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };

	glPushMatrix();
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, emit);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50.0);

	if (state->light0 == 0)
		glCallList(dl->emit_off_list);

	glTranslated(11.0, 15.9, 16.5);
	glCallList(dl->lights_list);
	glTranslated(-22.0, 0.0, 0.0);
	glCallList(dl->lights_list);
	glTranslated(0.0, 0.0, -16.5);
	glCallList(dl->lights_list);
	glTranslated(22.0, 0.0, 0.0);
	glCallList(dl->lights_list);
	glTranslated(0.0, 0.0, -16.5);
	glCallList(dl->lights_list);
	glTranslated(-22.0, 0.0, 0.0);
	glCallList(dl->lights_list);
	
	glCallList(dl->white_list);

	glPopMatrix();
}

/* Places lights in the scene */
void place_lights()
{
	//GLfloat light0_position[] = { 0.0, -1.0, 0.0, 0.0 };
	//GLfloat light0_ambient[] = { 0.8, 0.8, 0.8, 1.0 };
	GLfloat light2_position[] = { 12.0, 16.5, 16.5, 1.0 };
	GLfloat light3_position[] = { -12.0, 16.5, 16.5, 1.0 };
	GLfloat light4_position[] = { 12.0, 16.5, 0.0, 1.0 };
	GLfloat light5_position[] = { -12.0, 16.5, 0.0, 1.0 };
	GLfloat light6_position[] = { 12.0, 16.5, -16.5, 1.0 };
	GLfloat light7_position[] = { -12.0, 16.5, -16.5, 1.0 };
	GLfloat light_spot_direction[] = { 0.0, -1.0, 0.0 };
	GLfloat light_ambient[] = { 0.3, 0.3, 0.3, 1.0 };
	GLfloat light_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
	GLfloat light_specular[] = { 0.5, 0.5, 0.5, 1.0 };

	// fixed light
	/*
	glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_ambient);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	*/
	
	// six lights on the ceiling
	glLightfv(GL_LIGHT2, GL_POSITION, light2_position);
	glLightf(GL_LIGHT2, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT2, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT2, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT2, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT2);
	
	glLightfv(GL_LIGHT3, GL_POSITION, light3_position);
	glLightf(GL_LIGHT3, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT3, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT3, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT3, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT3);
	
	glLightfv(GL_LIGHT4, GL_POSITION, light4_position);
	glLightf(GL_LIGHT4, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT4, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT4, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT4, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT4, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT4, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT4, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT4);
	
	glLightfv(GL_LIGHT5, GL_POSITION, light5_position);
	glLightf(GL_LIGHT5, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT5, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT5, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT5, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT5, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT5, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT5, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT5);
	
	glLightfv(GL_LIGHT6, GL_POSITION, light6_position);
	glLightf(GL_LIGHT6, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT6, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT6, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT6, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT6, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT6, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT6, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT6);
	
	glLightfv(GL_LIGHT7, GL_POSITION, light7_position);
	glLightf(GL_LIGHT7, GL_SPOT_CUTOFF, 90.0);
	glLightfv(GL_LIGHT7, GL_SPOT_DIRECTION, light_spot_direction);
	glLightfv(GL_LIGHT7, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT7, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT7, GL_SPECULAR, light_specular);
	glLightf(GL_LIGHT7, GL_CONSTANT_ATTENUATION, 2.0);
	glLightf(GL_LIGHT7, GL_LINEAR_ATTENUATION, 0.05);
	glEnable(GL_LIGHT7);
}

/* Display callback function */
void display()
{
	GLenum error_code = GL_NO_ERROR;
	const GLubyte *error_string = NULL;
	object_t *op = obj_list;
	//object_t *nextop = NULL;
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// update camera vectors for display
//	reset_camera();
	update_camera();

	gluLookAt(cam->pos.x, cam->pos.y, cam->pos.z, cam->view.x,
		cam->view.y, cam->view.z, cam->up.x, cam->up.y, cam->up.z);

	place_lights();

	draw_hud();
	
	//glCallList(dl->white_list);

	draw_planet();
	
	glCallList(dl->window_list);
	
	glCallList(dl->walls_list);
	
	glCallList(dl->monitor_list);

	glPushMatrix();
	glRotated(180.0, 0.0, 1.0, 0.0);
	glTranslated(0.0, 0.0, -34.0);
	glCallList(dl->monitor_list);
	glPopMatrix();
	
	glCallList(dl->ceiling_list);
	
	glCallList(dl->floor_list);
	
	glCallList(dl->ventilation_list);

	glCallList(dl->decor_list);

	draw_lights();

	draw_console();

	glCallList(dl->white_list);
	
	// draw object list
	while (op != NULL)
	{
		if ((op->live > 0) && (op->draw))
		{
			glPushMatrix();
			glTranslated(op->translate_to.x, op->translate_to.y, op->translate_to.z);
			glRotated(op->rotate_around_x, 1.0, 0.0, 0.0);
			glRotated(op->rotate_around_y, 0.0, 1.0, 0.0);
			glRotated(op->rotate_around_z, 0.0, 0.0, 1.0);
			op->draw();
			glPopMatrix();
		}
		if ((left_shot->live || right_shot->live) && (op->live == 1))
		{
			if ((left_shot->live) && (distance(left_shot->vp,
			op->vp) < (left_shot->radius + (op->radius*5/6))))
			{
				if (op->on_collision)
					op->on_collision(op);
				add_score(100);
				left_shot->live = 0;
				left_shot->t = 0.0;
				explode(left_shot);
			}
			if ((right_shot->live) && (distance(right_shot->vp,
			op->vp) < (right_shot->radius + (op->radius*5/6))))
			{
				if (op->on_collision)
					op->on_collision(op);
				add_score(100);
				right_shot->live = 0;
				right_shot->t = 0.0;
				explode(right_shot);
			}
		}
		/*
		if (op->live == 0)
		{
			nextop = op->next;
			delete_object(op);
			op = nextop;
		}
		else
			op = op->next;
		*/
		op = op->next;
	}
	
	glCallList(dl->white_list);
	
	// shots and lights
	glDisable(GL_LIGHT1);
	lights_off();
	
	draw_explosions();
	
	if (state->light0 == 1)
		lights_on();
	if (state->light1)
		glEnable(GL_LIGHT1);
	
	draw_shots(left_shot);
	draw_shots(right_shot);

	draw_trails();
	
	glCallList(dl->white_list);

	// fence with alpha
	glCallList(dl->fence_list);

	// hud zone
	draw_compass();
	
	glCallList(dl->white_list);

	update_shooter();

	frame_counter();
	
	// check for errors
	if ((error_code = glGetError()) != GL_NO_ERROR)
	{
		error_string = gluErrorString(error_code);
		fprintf(stderr, "OpenGL error: %s\n", error_string);
	}

	glutSwapBuffers();
}
