/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// lib.h
// a simple 3D vector library

#ifndef ALPSHOOT_LIB_H
#define ALPSHOOT_LIB_H

#include <math.h>
#include "defs.h"
#include "camera.h"

GLdouble float_abs(GLdouble f);
//double calculateAngle(double size, double distance);
GLdouble add_angle(GLdouble angle, GLdouble delta);
// matrix operations
void identity_matrix(GLdouble m[16]);
void apply_v2m(vector *v, GLdouble m[16]);
void rotation_matrix_X(GLdouble angle, GLdouble m[16]);
void rotation_matrix_Y(GLdouble angle, GLdouble m[16]);
void rotation_matrix_Z(GLdouble angle, GLdouble m[16]);
void rotation_matrix_view(GLdouble angle, GLdouble m[16]);
void matrix_multiply(GLdouble m[16], GLdouble n[16], GLdouble r[16]);
void rotation_matrix(GLdouble angle, vector v, GLdouble m[16]);
void translation_matrix(GLdouble dx, GLdouble dy, GLdouble dz, GLdouble m[16]);
void print_vector(vector x, const char *name);
void print_matrix(GLdouble m[16]);
void orthonormalize_matrix(GLdouble m[16]);
void matrix_to_angles(GLdouble m[16], GLdouble *angleX, GLdouble *angleY, GLdouble *angleZ);
// vector operations
void copy_vector(vector s, vector *d);
GLuint vec_equal(vector u, vector v);
vector vec_subtract(vector v, vector u);
vector vec_add(vector v, vector u);
vector vec_divide_scalar(vector v, GLdouble scalar);
vector vec_multiply_scalar(vector v, GLdouble scalar);
vector vec_negate(vector v);
GLdouble magnitude(vector v);
vector normalize(vector v);
GLdouble dot_product(vector v, vector u);
vector cross_product(vector v, vector u);
vector normal(vector v1, vector v2, vector v3);
GLdouble angle_between_vectors(vector v, vector u);
GLdouble distance(vector v, vector u);
void rotate(GLdouble angle, vector axis1, vector *axis2, vector *axis3);
void translate(GLdouble dx, GLdouble dy, GLdouble dz, vector *v);
// point and planes operations
vector point_on_curve(vector u, vector v, GLdouble t);
vector point_on_curve2(vector u, vector v, vector w, vector y, GLdouble t);
vector de_cast2(vector v1, vector v2, GLdouble t);
vector de_cast4(vector v1, vector v2, vector v3, vector v4, GLdouble t);
vector closest_point_on_line(vector v1, vector v2, vector p);
GLdouble plane_distance(vector n, vector p);
GLint intersected_plane(vector v1, vector v2, vector v3, vector p1, vector p2, vector *norm, GLdouble *dist);
vector intersection_point(vector norm, vector p1, vector p2, GLdouble dist);
GLint inside_polygon(vector intersect, vector poly[], long v_count);
vector find_rotated_coords(vector vec, GLdouble angle, GLdouble x, GLdouble y, GLdouble z);

#endif
