/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// particles.h
// minimal particles engine header file
// some code here is public domain code by: Gustav Taxen, nv91-gta@nada.kth.se

#ifndef ALPSHOOT_PARTICLES_H
#define ALPSHOOT_PARTICLES_H

#define NUM_PARTICLES	1000
#define NUM_DEBRIS	800
#define NUM_TRAIL	5

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include "defs.h"

/* a particle */
struct particle_data
{
	float position[3];
	float speed[3];
	float color[4];
};
typedef struct particle_data particle_data;

/* a piece of debris */
struct debris_data
{
	float position[3];
	float speed[3];
	float orientation[3]; // rotation angles around x, y, and z axes
	float orientationSpeed[3];
	float color[4];
	float scale[3];
};
typedef struct debris_data debris_data;

/* explosion structure */
struct explosion_t
{
	particle_data particles[NUM_PARTICLES];
	debris_data debris[NUM_DEBRIS];
	int fuel;
	struct explosion_t *next;
	struct explosion_t *prev;
};
typedef struct explosion_t explosion_t;

/* rocket trail structure */
struct trail_t
{
	particle_data particles[NUM_TRAIL];
	int fuel;
	struct trail_t *next;
	struct trail_t *prev;
};
typedef struct trail_t trail_t;

explosion_t *add_explosion();
void delete_explosion(explosion_t *o);
trail_t *add_trail();
void delete_trail(trail_t *o);
void new_speed(GLfloat dest[3], int wantNormalize);
void new_explosion(int wantNormalize, vector vp);
void new_trail(int wantNormalize, vector vp);

extern explosion_t *exp_list;
extern explosion_t *last_exp;
extern trail_t *trail_list;
extern trail_t *last_trail;

#endif
