/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// world.h
// world manipulation functions header

#ifndef ALPSHOOT_WORLD_H
#define ALPSHOOT_WORLD_H

#ifdef ALPSHOOT_WIN32
#include <windows.h>
#endif

#ifndef ALPSHOOT_MACOSX
#include <GL/glut.h>
#else
#include <glut.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "lib.h"
#include "quatlib.h"

/* list of objects to draw */
struct object_t {
	// current object position
	vector vp;
	// initial object position (for moving objects)
	vector init_vp;
	// extra control points (for some moving objects)
	vector cp1;
	vector cp2;
	// final object position (for moving objects)
	vector final_vp;
	// bounding box radius
	GLdouble radius;
	// angles to rotate around axes
	GLdouble rotate_around_x;
	GLdouble rotate_around_y;
	GLdouble rotate_around_z;
	// orientation quaternion for this object
	quaternion orientation;
	// vector used to translate object from origin
	vector translate_to;
	// pointer to function to draw this object
	void (*draw)(void);
	// pointer to collision management function
	void (*on_collision)(struct object_t*);
	// object is drawn if live flag != 0
	GLint live;
	// t is used for moving objects (eg. shots)
	GLdouble t;
	// alpha value used for some effects
	GLdouble alpha;
	// pointer to next element
	struct object_t *next;
	// pointer to previous element
	struct object_t *prev;
};
typedef struct object_t object_t;

object_t* add_object();
void insert_object(object_t* o);
void delete_object(object_t* p);
object_t *make_object(GLdouble r, GLdouble rotx, GLdouble roty, GLdouble rotz,
GLdouble tx, GLdouble ty, GLdouble tz, void (*drawfun)(void), void (*collidefun)(object_t*));
void add_score(GLint num);

extern object_t *obj_list, *last_obj;

#endif
